"""Master notebook transformer for Otter Assign"""

import copy
import fica
import nbformat as nbf
import pathlib

from typing import Union

from .assignment import Assignment
from .blocks import BlockType, get_cell_config, is_assignment_config_cell, is_block_boundary_cell
from .cell_factory import CellFactory
from .feature_toggle import FeatureToggle
from .plugins import replace_plugins_with_calls
from .question_config import QuestionConfig
from .r_adapter import rmarkdown_converter
from .r_adapter.cell_factory import RCellFactory
from .solutions import (
    ANSWER_CELL_TAG,
    has_seed,
    overwrite_seed_vars,
    SOLUTION_CELL_TAG,
    strip_ignored_lines,
    strip_solutions_and_output,
)
from .tests_manager import AssignmentTestsManager
from .utils import (
    add_tag,
    AssignNotebookFormatException,
    is_cell_type,
    is_ignore_cell,
    lock,
    remove_cell_ids_if_applicable,
)
from ..nbmeta_config import NBMetadataConfig
from ..utils import get_source, NOTEBOOK_METADATA_KEY


class NotebookTransformer:
    """
    A transformer for master notebooks that converts them to autograder- and
    student-formatted notebooks.

    A note on terminology: An *autograder-formatted* notebook is the same as the master notebook but
    with all metadata and structural cells removed, and cells that set up and use Otter's client
    package (e.g. init cells, check cells, etc.) added. A *student-formatted* notebook is the same
    as an autograder-formatted notebook but has had its solutions and hidden tests stripped. (The
    conversion from an autograder-formatted notebook to a student-formatted notebook is not actually
    handled in this class but in the TransformedNotebookContainer class.)

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment config
        tests_mgr (``otter.assign.tests_manager.AssignmentTestsManager``): the tests manager to
            populate
    """

    assignment: Assignment
    """the assignment config"""

    cell_factory: CellFactory
    """a factory for various kinds of cells for using Otter's client package"""

    tests_mgr: AssignmentTestsManager
    """the tests manager to populate"""

    def __init__(self, assignment: Assignment, tests_mgr: AssignmentTestsManager):
        self.assignment = assignment
        self.cell_factory = (RCellFactory if self.assignment.is_r else CellFactory)(self.assignment)
        self.tests_mgr = tests_mgr

    def add_export_tag_to_cell(self, cell: nbf.NotebookNode, end: bool = False) -> nbf.NotebookNode:
        """
        Add an HTML comment to open or close question export for PDF filtering to the top of
        ``cell``. ``cell`` should be a Markdown cell.

        Args:
            cell (``nbformat.NotebookNode``): the cell to add the close export to

        Returns:
            ``nbformat.NotebookNode``: a copy of ``cell`` with the close export comment at the top
        """
        if not FeatureToggle.PDF_FILTERING_COMMENTS.value.is_enabled(self.assignment):
            return cell

        cell = copy.deepcopy(cell)
        source = get_source(cell)
        tag = "<!-- " + ("END" if end else "BEGIN") + " QUESTION -->"
        source = [tag, ""] + source
        cell["source"] = "\n".join(source)
        lock(cell)
        return cell

    @staticmethod
    def add_point_value_info_to_cell(
        cell: nbf.NotebookNode,
        points: Union[int, float],
    ) -> nbf.NotebookNode:
        """
        Add the point value information to the provided cell, returning a copy.

        Args:
            cell (``nbformat.NotebookNode``): the cell to add to
            points (``int | float``): the point value of the question

        Returns:
            ``nbformat.NotebookNode``: a copy of ``cell`` with the point value
        """
        cell = copy.deepcopy(cell)
        source = get_source(cell)
        source.extend(["", f"_Points:_ {points}"])
        cell["source"] = "\n".join(source)
        return cell

    def transform_notebook(self, nb: nbf.NotebookNode) -> "TransformedNotebookContainer":
        """
        Transform a master notebook into an autograder-formatted notebook.

        Args:
            nb (``nbformat.NotebookNode``): the master notebook

        Returns:
            ``TransformedNotebookContainer``: the autograder-formatted notebook
        """
        transformed_cells = self._get_transformed_cells(nb["cells"])

        if self.assignment.init_cell:
            transformed_cells = self.cell_factory.create_init_cells() + transformed_cells

        if self.assignment.check_all_cell:
            transformed_cells += self.cell_factory.create_check_all_cells()

        if self.assignment.export_cell:
            transformed_cells += self.cell_factory.create_export_cells()

        transformed_nb = copy.deepcopy(nb)
        transformed_nb["cells"] = transformed_cells

        # replace plugins
        transformed_nb = replace_plugins_with_calls(transformed_nb)

        # strip out ignored lines
        transformed_nb = strip_ignored_lines(transformed_nb)

        remove_cell_ids_if_applicable(transformed_nb)

        return TransformedNotebookContainer(transformed_nb, self)

    def _get_transformed_cells(self, cells: list[nbf.NotebookNode]) -> list[nbf.NotebookNode]:
        """
        Takes in a list of cells from the master notebook and returns a list of cells for the
        autograder notebook.

        Args:
            cells (``list[nbf.NotebookNode]``): the cells of the master notebook

        Returns:
            ``list[nbf.NotebookNode]``: the cells of the autograder notebook
        """
        curr_block: list[BlockType] = []  # allow nested blocks
        transformed_cells: list[nbf.NotebookNode] = []

        question, has_prompt, no_solution, last_question_md_cell = None, False, False, None

        solution_has_md_cells, prompt_insertion_index = False, None

        need_begin_export, need_end_export = False, False

        for i, cell in enumerate(cells):
            if is_ignore_cell(cell):
                continue

            # check for assignment config
            if is_assignment_config_cell(cell):
                config = get_cell_config(cell)
                if config.get("config_file"):
                    self.assignment.load_config_file(config["config_file"])
                self.assignment.update(config)
                if self.assignment.runs_on == "colab" and not self.assignment.tests.files:
                    raise ValueError(
                        "notebooks that run on Colab must use test files and not notebook metadata tests"
                    )
                continue

            # check for an end to the current block
            if len(curr_block) > 0 and is_block_boundary_cell(cell, curr_block[-1], end=True):
                block_type = curr_block.pop()  # remove the block

                if block_type is BlockType.QUESTION:

                    if question.manual or question.export:
                        need_end_export = True

                    # generate a check cell if necessary
                    if self.tests_mgr.has_tests(question):
                        check_cells = self.cell_factory.create_check_cells(question)

                        # only add to notebook if there's a response cell or if there are public tests;
                        # don't add cell if the 'check_cell' key of question is false
                        if (
                            not no_solution or self.tests_mgr.any_public_tests(question)
                        ) and question.check_cell:
                            transformed_cells.extend(check_cells)
                        else:
                            last_cell_meta = transformed_cells[-1]["metadata"]
                            last_cell_meta.setdefault(NOTEBOOK_METADATA_KEY, {})
                            last_cell_meta[NOTEBOOK_METADATA_KEY]["tests"] = [question.name]

                    # add points to question cell if specified
                    if self.assignment.show_question_points and last_question_md_cell is not None:
                        points = self.tests_mgr.determine_question_point_value(question)
                        transformed_cells[last_question_md_cell] = (
                            self.add_point_value_info_to_cell(
                                transformed_cells[last_question_md_cell], points
                            )
                        )

                    question, has_prompt, no_solution, last_question_md_cell = (
                        None,
                        False,
                        False,
                        None,
                    )
                    solution_has_md_cells, prompt_insertion_index = False, None

                elif block_type is BlockType.SOLUTION:
                    if not has_prompt and solution_has_md_cells and question.manual:
                        if prompt_insertion_index is None:
                            raise RuntimeError("Could not find prompt insertion index")
                        transformed_cells.insert(
                            prompt_insertion_index, CellFactory.create_markdown_response_cell()
                        )
                        has_prompt = True

                continue  # if this is an end to the last nested block, we're OK

            # check for invalid block ends
            for block_type in BlockType:
                if is_block_boundary_cell(cell, block_type, end=True):

                    # if a child is missing an end block cell, raise an error
                    if block_type in curr_block:
                        raise AssignNotebookFormatException(
                            f"Found an end {block_type.value} cell with an un-ended child "
                            + f"{curr_block[-1].value} block",
                            question,
                            i,
                        )

                    # otherwise raise an error for an end with no begin
                    else:
                        raise AssignNotebookFormatException(
                            f"Found an end {block_type.value} cell with no begin block cell",
                            question,
                            i,
                        )

            # check for begin blocks
            found_begin = False
            for block_type in BlockType:
                if is_block_boundary_cell(cell, block_type):
                    found_begin = True
                    break

            if found_begin:
                if len(curr_block) == 0 and block_type is not BlockType.QUESTION:
                    raise AssignNotebookFormatException(
                        f"Found a begin {block_type.value} cell outside a question", question, i
                    )
                elif len(curr_block) > 0 and block_type is BlockType.QUESTION:
                    raise AssignNotebookFormatException(
                        f"Found a begin {block_type.value} cell inside another question",
                        question,
                        i,
                    )
                elif len(curr_block) > 1:
                    raise AssignNotebookFormatException(
                        f"Found a begin {block_type.value} cell inside a {curr_block[-1].value} "
                        "block",
                        question,
                        i,
                    )
                elif block_type is BlockType.PROMPT and has_prompt:
                    # has_prompt was set by the solution block
                    raise AssignNotebookFormatException(
                        "Found a prompt block after a solution block", question, i
                    )

                # if not an invalid begin cell, update state
                if block_type is BlockType.PROMPT:
                    has_prompt = True

                elif block_type is BlockType.SOLUTION and not has_prompt:
                    prompt_insertion_index = len(transformed_cells)

                elif block_type is BlockType.TESTS and not has_prompt:
                    no_solution = True

                elif block_type is BlockType.QUESTION:
                    question_config = get_cell_config(cell)
                    if not isinstance(question_config, dict):
                        raise AssignNotebookFormatException(
                            "Found a begin question cell with no config", None, i
                        )

                    question = QuestionConfig(question_config)
                    self.tests_mgr.add_question(question)
                    if question.manual or question.export:
                        need_begin_export = True

                curr_block.append(block_type)
                continue

            # if in a block, process the current cell
            if len(curr_block) > 0:
                if curr_block[-1] == BlockType.TESTS:
                    if not is_cell_type(cell, "code"):
                        raise AssignNotebookFormatException(
                            "Found a non-code cell in tests block", question, i
                        )
                    self.tests_mgr.read_test(cell, question)
                    continue

                elif curr_block[-1] == BlockType.PROMPT:
                    cell = add_tag(cell, ANSWER_CELL_TAG)

                elif curr_block[-1] == BlockType.SOLUTION:
                    cell = add_tag(cell, SOLUTION_CELL_TAG)

                    if is_cell_type(cell, "markdown"):
                        solution_has_md_cells = True

                elif curr_block[-1] == BlockType.QUESTION and is_cell_type(cell, "markdown"):
                    last_question_md_cell = len(transformed_cells)

            # add export tags if needed
            export_delim_cell = None
            if need_begin_export:
                if is_cell_type(cell, "markdown"):
                    cell = self.add_export_tag_to_cell(cell)
                elif FeatureToggle.PDF_FILTERING_COMMENTS.value.is_enabled(self.assignment):
                    export_delim_cell = nbf.v4.new_markdown_cell()
                    export_delim_cell = self.add_export_tag_to_cell(export_delim_cell)
                need_begin_export = False
            if need_end_export:
                if is_cell_type(cell, "markdown"):
                    cell = self.add_export_tag_to_cell(cell, end=True)
                elif FeatureToggle.PDF_FILTERING_COMMENTS.value.is_enabled(self.assignment):
                    if export_delim_cell is None:
                        export_delim_cell = nbf.v4.new_markdown_cell()
                    export_delim_cell = self.add_export_tag_to_cell(export_delim_cell, end=True)
                need_end_export = False

            if export_delim_cell is not None:
                transformed_cells.append(export_delim_cell)

            if has_seed(cell):
                self.assignment.seed_required = True

            # if this cell is part of the question text, lock it
            if len(curr_block) == 1:
                lock(cell)

            # this is just a normal cell so add it to transformed_cells
            transformed_cells.append(cell)

        # if the last cell was the end of a manually-graded question, add a close export tag
        if need_end_export:
            transformed_cells.append(
                self.add_export_tag_to_cell(nbf.v4.new_markdown_cell(), end=True)
            )

        return transformed_cells


class TransformedNotebookContainer:
    """
    A container for the autograder notebook returned by ``NotebookTransformer.transform_notebook``
    that contains logic for writing the notebook file (both autograder and student versions) and
    the test files.

    Args:
        transformed_nb (``nbformat.NotebookNode``): the autograder notebook
        nb_transformer (``NotebookTransformer``): the notebook transformer used to create
            ``transformed_nb``
    """

    transformed_nb: nbf.NotebookNode
    """the autograder notebook"""

    nb_transformer: NotebookTransformer
    """the notebook transformer used to create ``transformed_nb``"""

    nbmeta_config: NBMetadataConfig
    """the notebook metadata config to include when writing the notebook"""

    def __init__(self, transformed_nb: nbf.NotebookNode, nb_transformer: NotebookTransformer):
        self.transformed_nb = transformed_nb
        self.nb_transformer = nb_transformer

        # populate NBMetadataConfig
        self.nbmeta_config = NBMetadataConfig({})
        if self.assignment.name:
            self.nbmeta_config.assignment_name = self.assignment.name
        if self.assignment.export_cell and self.assignment.export_cell.require_no_pdf_ack:
            self.nbmeta_config.require_no_pdf_confirmation = True
            if isinstance(self.assignment.export_cell.require_no_pdf_ack, fica.Config):
                self.nbmeta_config.export_pdf_failure_message = (
                    self.assignment.export_cell.require_no_pdf_ack.message
                )

    @property
    def assignment(self) -> Assignment:
        """the assignment config for this notebook"""
        return self.nb_transformer.assignment

    def _get_sanitized_nb(self) -> nbf.NotebookNode:
        """
        Return a copy of ``self.transformed_nb`` with solutions and outputs stripped and seed
        variables replaced.

        Returns:
            ``nbformat.NotebookNode``: the student notebook
        """
        nb = strip_solutions_and_output(self.assignment, self.transformed_nb)
        if self.nb_transformer.assignment.seed.variable:
            nb = overwrite_seed_vars(
                nb,
                self.nb_transformer.assignment.seed.variable,
                self.nb_transformer.assignment.seed.student_value,
            )
        return nb

    def _add_nbmeta_config(self, nb: nbf.NotebookNode):
        """
        Add the notebook metadata config to the provided notebook's metadata in-place.
        """
        uc = self.nbmeta_config.get_user_config()
        if uc:
            nb["metadata"][NOTEBOOK_METADATA_KEY] = uc

    def write_transformed_nb(self, output_path: Union[pathlib.Path, str], sanitize: bool):
        """
        Write the transformed notebook (either as an autograder or student notebook) to the
        specified file, converting to R Markdown if necessary.

        Args:
            output_path (``pathlib.Path | str``): the path at which to write the notebook
            sanitize (``bool``): whether to sanitize the notebook (i.e. write the student version)
        """
        nb = self._get_sanitized_nb() if sanitize else self.transformed_nb
        self._add_nbmeta_config(nb)

        if self.nb_transformer.assignment.is_rmd:
            rmarkdown_converter.write_as_rmd(nb, str(output_path), not sanitize)

        else:
            try:
                from nbformat.validator import normalize
            except ImportError:
                normalize = lambda nb: (0, nb)

            _, nb = normalize(nb)
            nbf.write(nb, str(output_path))

    def write_tests(
        self,
        tests_dir: Union[pathlib.Path, str, None],
        include_hidden: bool,
        force_files: bool,
    ):
        """
        Write the tests, either to the notebook metadata or files in ``tests_dir`` as indicated by
        the assignment config.

        Args:
            tests_dir (``pathlib.Path | str | None``): the path to the directory of tests, if any
            include_hidden (``bool``): whether to include hidden test cases
            force_files (``bool``): whether to force writing the tests as files (overriding the
                assignment config)
        """
        self.nb_transformer.tests_mgr.write_tests(
            self.nbmeta_config,
            tests_dir,
            include_hidden=include_hidden,
            force_files=force_files,
        )



"""Cell generators for R notebooks"""

import nbformat

from ..cell_factory import CellFactory
from ..feature_toggle import FeatureToggle
from ..utils import lock


class RCellFactory(CellFactory):
    """
    A cell factory for R assignments.
    """

    def create_init_cells(self):
        return []

    def create_check_cells(self, question):
        cell = nbformat.v4.new_code_cell()
        cell.source = ['. = ottr::check("tests/{}.R")'.format(question.name)]
        lock(cell)
        return [cell]

    def create_check_all_cells(self):
        return []

    def create_export_cells(self):
        if not self.assignment.export_cell:
            return []

        force_save = not self.assignment.is_rmd and self.assignment.export_cell.force_save

        instructions = nbformat.v4.new_markdown_cell()
        instructions.source = (
            "## Submission\n\nMake sure you have run all cells in your "
            "notebook in order before running the cell below, so that all images/graphs appear "
            "in the output. The cell below will generate a zip file for you to submit."
        )

        if not force_save:
            instructions.source += " **Please save before exporting!**"

        if self.assignment.export_cell.instructions:
            instructions.source += "\n\n" + self.assignment.export_cell.instructions

        args = []
        if self.assignment.export_cell.pdf:
            args.append("pdf = TRUE")
        if force_save:
            args.append("force_save = TRUE")

        if args:
            args = ", " + ", ".join(args)
        else:
            args = ""

        export = nbformat.v4.new_code_cell()
        source_lines = (
            []
            if force_save
            else ["# Save your notebook first, then run this cell to export your submission."]
        )
        source_lines.append(f'ottr::export("{self.assignment.notebook_basename}"{args})')
        export.source = "\n".join(source_lines)

        lock(instructions)
        lock(export)

        cells = [instructions, export]
        if self.check_feature_toggle(FeatureToggle.EMPTY_MD_BOUNDARY_CELLS):
            cells.append(nbformat.v4.new_markdown_cell(" "))  # add buffer cell

        return cells



"""ABC for Otter Export exporters"""

import importlib.resources
import nbformat

from abc import ABC, abstractmethod
from typing import Any

from . import __name__ as pkg_name
from .utils import has_begin, has_end, sub_end_for_new_page
from ...utils import NBFORMAT_VERSION


TEMPLATE_DIR = importlib.resources.files(pkg_name) / "templates"


class ExportFailedException(Exception):
    """
    A generic error class for Otter Export to raise when an export fails.
    """


class BaseExporter(ABC):
    """
    Abstract base class for Otter Export exporters

    Defines methods that read in notebooks from file paths and filter cells with pagebreak annotations
    if necessary. Sublcasses are not meant to be instatiated but have class methods called. The abstract
    class method ``BaseExporter.convert_notebook`` does the heavy lifting of converting a notebook
    file path into a PDF file.

    Attributes:
        default_options (``dict``): the default options for all exporters
    """

    default_options = {
        "filtering": False,
        "pagebreaks": True,
    }

    @classmethod
    @abstractmethod
    def convert_notebook(cls, nb_path: str, dest: str, **kwargs: Any):
        """
        Writes a notebook at ``nb_path`` to a PDF file

        Args:
            nb_path (``str``): path to notebook
            dest (``str``): path to write PDF
            **kwargs: additional arguments use during conversion by subclasses
        """
        ...

    @classmethod
    def load_notebook(
        cls, nb_path: str, filtering: bool = False, pagebreaks: bool = True
    ) -> nbformat.NotebookNode:
        """
        Loads notebook at ``nb_path`` with nbformat and returns the parsed notebookly filtered
        and with pagebreak metadata hidden in HTML comments.

        Args:
            nb_path (``str``): path to notebook
            filtering (``bool``): whetheer cells should be filtered
            pagebreaks (``bool``): whether to include pagebreaks between each question; ignored
                if ``filtering`` is ``False``

        Returns:
            ``nbformat.NotebookNode``: the parsed and (optionally) filtered notebook
        """
        with open(nb_path) as f:
            notebook = nbformat.read(f, as_version=NBFORMAT_VERSION)
        if filtering:
            notebook = cls.filter_cells(notebook, pagebreaks=pagebreaks)
        return notebook

    @classmethod
    def filter_cells(cls, notebook: nbformat.NotebookNode, pagebreaks: bool = True):
        """
        Filters a parsed notebook using HTML comments in Markdown cells. Optionally inserts pagebreak
        metadata as HTML comments.

        Args:
            notebook (``nbformat.NotebookNode``): the parsed notebook
            pagebreaks (``bool``): whether to include pagebreaks between questions

        Returns:
            ``nbformat.NotebookNode``: the filtered notebook with (optional) pagebreaks
        """
        cells = notebook["cells"]

        idx_to_delete, in_question = [], False
        for curr_idx, cell in enumerate(cells):

            if isinstance(cell["source"], str):
                source = cell["source"].split("\n")
            elif isinstance(cell["source"], list):
                source = [line.strip("\n") for line in cell["source"]]
            else:
                raise Exception("Invalid notebook cell source: {}".format(cell))

            lines_before_begin, lines_after_end = -1, -1
            for line_idx, line in enumerate(source):

                # check for begin question regex in source
                if not in_question:
                    if has_begin(line) and lines_before_begin == -1:
                        lines_before_begin = line_idx
                        in_question = True

                # check for end question regex in source
                else:
                    if has_end(line):
                        lines_after_end = line_idx
                        in_question = False

                        # if we are creating pagebreaks, sub the end regex for a newpage directive
                        if pagebreaks:
                            source[line_idx] = sub_end_for_new_page(line)

            # if we are not in question and there is no begin/end comment, delete the cell
            if lines_before_begin == -1 and lines_after_end == -1 and not in_question:
                idx_to_delete.append(curr_idx)

            # if both are in cell and before is after end, delete intervening lines
            elif (
                lines_before_begin != -1
                and lines_after_end != -1
                and lines_after_end < lines_before_begin
            ):
                del source[lines_after_end + 1 : lines_before_begin]

            else:
                # if there is an end comment, delete lines after that
                if lines_after_end != -1:
                    del source[lines_after_end + 1 :]

                # if there is a begin comment, delete lines before that
                if lines_before_begin != -1:
                    del source[:lines_before_begin]

            # update source
            cell["source"] = "\n".join(source)

        # reverse indices list so that we do not need to decrement while deleting
        idx_to_delete.reverse()

        # delete at indices
        for idx in idx_to_delete:
            del cells[idx]

        return notebook



import logging
import logging.handlers

from contextlib import contextmanager
from multiprocessing import Queue
from typing import Optional


ERROR = logging.ERROR
WARNING = logging.WARNING
INFO = logging.INFO
DEBUG = logging.DEBUG


_instances: dict[str, logging.Logger] = {}
"""all ``logging.Logger`` instances by name"""

_log_level: int = logging.WARNING
"""the current log level"""

_handler = logging.StreamHandler()
"""the formatter to use when formatting log records"""

_handlers: list[logging.Handler] = [_handler]
"""extra handlers that should receive logs"""


# set a formatter on _handler
_handler.setFormatter(logging.Formatter("[%(levelname)s %(name)s.%(funcName)s] %(message)s"))


def send_logs(host: str, port: int):
    """
    Add a ``SocketHandler`` to all loggers that sends their logs to a TCP socket at the
    specified host and port.
    """
    socket_handler = logging.handlers.SocketHandler(host, port)
    _handlers.append(socket_handler)
    for logger in _instances.values():
        logger.addHandler(socket_handler)


def get_logger(name: str) -> logging.Logger:
    """
    Retrieve ``logging.Logger`` with name ``name`` and return it, setting the log level to the
    class log level.
    """
    if name in _instances:
        return _instances[name]

    logger = logging.getLogger(name)
    logger.propagate = False  # prevent child loggers from inheriting the handler
    logger.setLevel(_log_level)

    for h in _handlers:
        logger.addHandler(h)

    _instances[name] = logger
    return logger


def _get_queue_handler() -> Optional["QueueLoggingHandler"]:
    for h in _handlers:
        if isinstance(h, QueueLoggingHandler):
            return h
    return None


def add_queue_handler(result_queue: "Queue[str]"):
    """
    Set up a ``QueueLoggingHandler`` that sends all logged messages to the provided
    ``multiprocessing.Queue``.

    Args:
        result_queue (``multiprocessing.Queue[str]``): the queue to write logs to
    """
    if _get_queue_handler() is not None:
        raise RuntimeError("A QueueLoggingHandler is already set up")

    queue_handler = QueueLoggingHandler(result_queue)
    _handlers.append(queue_handler)
    for logger in _instances.values():
        logger.addHandler(queue_handler)


def remove_queue_handlers():
    """
    Remove the current ``QueueLoggingHandler`` if one is set up.
    """
    qh = _get_queue_handler()
    if qh is None:
        return
    for logger in _instances.values():
        logger.removeHandler(qh)
    _handlers.remove(qh)


def get_level() -> int:
    """
    Return the current log level of these loggers.
    """
    return _log_level


def set_level(log_level: int):
    """
    Set the log levels for all ``Logger``s created by this class (existing and future).
    """
    global _log_level
    _log_level = log_level
    for logger in _instances.values():
        logger.setLevel(log_level)


@contextmanager
def level_context(log_level: int):
    """
    Set the log level to a new value temporarily in a context.
    """
    curr_level = get_level()
    set_level(log_level)
    yield
    set_level(curr_level)


def reset_level():
    """
    Set the log levels for all ``Loggers`` created by this class (existing and future) back to
    ``logging.WARNING``.
    """
    set_level(logging.WARNING)


class Loggable:
    """
    A class for inheriting from that provides a logger via a class- and instance-accessible field.
    """

    _logger_instance: Optional[logging.Logger] = None

    @classmethod
    def _load_logger(cls):
        """
        Set-up the ``_logger`` field.
        """
        if cls._logger_instance is None:
            name = cls.__module__ + "." + cls.__name__
            cls._logger_instance = get_logger(name)

    @property
    def _logger(self) -> logging.Logger:
        """
        ``logging.Logger``: the logger instance for this class
        """
        return self._get_logger()

    @classmethod
    def _get_logger(cls) -> logging.Logger:
        """
        Load and return the logger for this class.

        Returns:
            ``logging.Logger``: the logger instance for this class
        """
        cls._load_logger()
        return cls._logger_instance


class QueueLoggingHandler(logging.Handler):
    """The logging handler that writes INFO messages to the log_queue.

    Args:
        log_queue (``multiprocessing.Queue``): the queue this handler writes to
    """

    def __init__(self, log_queue: "Queue[str]"):
        super().__init__()
        self.log_queue = log_queue

    def emit(self, record: logging.LogRecord):
        try:
            log_entry = self.format(record)
            self.log_queue.put(log_entry)
        except Exception:
            self.handleError(record)



"""Version and printable logo"""

import importlib
import sys

from textwrap import dedent, indent


__version__ = "6.1.6"


LOGO_WITH_VERSION = rf"""
  _________        __          __               
 /  _____  \    __|  |__    __|  |__               
|  /     \  |  |__    __|  |__    __|   _______    _  _____
| |       | |     |  |        |  |     |  ___  |  | |/ ____|
| |       | |     |  |        |  |     | |___| |  |   /    
| |       | |     |  |        |  |     | ______|  |  |
|  \_____/  |     |  |_       |  |_    | |_____   |  |
 \_________/       \ __|       \ __|    \______|  |__|
                                                v{__version__}
"""[
    1:
]  # remove beginning newline


_ADDITIONAL_PACKAGES = [
    "dill",
    "fica",
    "IPython",
    "nbconvert",
    "nbformat",
]


def print_version_info(logo: bool = False):
    """
    Prints the Otter logo and version information to stdout

    Args:
        logo (``bool``): whether to print the logo
    """
    if logo:
        print(LOGO_WITH_VERSION)
    print(
        dedent(
            f"""\
                Python version: {".".join(str(v) for v in sys.version_info[:3])}
                Otter-Grader version: {__version__}

                Additional package versions:
            """
        )
        + indent(_list_addl_package_versions(), "    ")
    )


def _list_addl_package_versions() -> str:
    """Generate a string listing versions of additional packages of interest."""
    versions = []
    for p in _ADDITIONAL_PACKAGES:
        try:
            m = importlib.import_module(p)
            versions.append(f"{p}: {m.__version__}")
        except Exception as e:
            versions.append(f"{p}: (failed) {type(e).__name__}: {e}")
    return "\n".join(versions)
